"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json


gif_response = json.loads("""{
  "results": [
    {
      "created": 1520636640.424062, 
      "url": "https://tenor.com/VDng.gif", 
      "media": [
        {
          "nanomp4": {
            "url": "https://media.tenor.com/videos/0cb9dfebd15018a5c41c4b09e38f4603/mp4", 
            "dims": [
              76, 
              40
            ], 
            "duration": 8.933333, 
            "preview": "https://media.tenor.com/images/731e5fe36ee192422482edea66fc9620/tenor.png", 
            "size": 68508
          }, 
          "nanowebm": {
            "url": "https://media.tenor.com/videos/3810dce6869fe053a733a94da6bbfe76/webm", 
            "dims": [
              76, 
              40
            ], 
            "preview": "https://media.tenor.com/images/731e5fe36ee192422482edea66fc9620/tenor.png", 
            "size": 73978
          }, 
          "tinygif": {
            "url": "https://media.tenor.com/images/6dae491b16b00707e0658c568b93bdab/tenor.gif", 
            "dims": [
              220, 
              118
            ], 
            "preview": "https://media.tenor.com/images/9caf27374675050d737a36b7095ee2c3/tenor.gif", 
            "size": 456504
          }, 
          "tinymp4": {
            "url": "https://media.tenor.com/videos/d406e80cb3498626f6021205051feb2b/mp4", 
            "dims": [
              166, 
              88
            ], 
            "duration": 8.933333, 
            "preview": "https://media.tenor.com/images/f3ed024334277a6d9222d9d8013b8780/tenor.png", 
            "size": 166950
          }, 
          "tinywebm": {
            "url": "https://media.tenor.com/videos/8e0b912e941d221deefa2d86de201969/webm", 
            "dims": [
              166, 
              88
            ], 
            "preview": "https://media.tenor.com/images/f3ed024334277a6d9222d9d8013b8780/tenor.png", 
            "size": 217640
          }, 
          "webm": {
            "url": "https://media.tenor.com/videos/d6ced731c3461d084e057421efb81fe0/webm", 
            "dims": [
              416, 
              224
            ], 
            "preview": "https://media.tenor.com/images/1b7446613a50219636b4299992e6daf0/tenor.png", 
            "size": 237925
          }, 
          "gif": {
            "url": "https://media.tenor.com/images/ad84b1269b576e0d87b74a8adc09295c/tenor.gif", 
            "dims": [
              416, 
              224
            ], 
            "preview": "https://media.tenor.com/images/1b7446613a50219636b4299992e6daf0/tenor.png", 
            "size": 7655329
          }, 
          "mp4": {
            "url": "https://media.tenor.com/videos/1a99850da5f5b278d394e0da3667828a/mp4", 
            "dims": [
              416, 
              224
            ], 
            "duration": 8.933333, 
            "preview": "https://media.tenor.com/images/1b7446613a50219636b4299992e6daf0/tenor.png", 
            "size": 396737
          }, 
          "nanogif": {
            "url": "https://media.tenor.com/images/39cc928d325dd3ee6ab9089121ed97a0/tenor.gif", 
            "dims": [
              167, 
              90
            ], 
            "preview": "https://media.tenor.com/images/216fb17dd2fb92d860eaecac97626741/tenor.gif", 
            "size": 281852
          }, 
          "mediumgif": {
            "url": "https://media.tenor.com/images/989f5142a249e25c776e3d47682a8bd9/tenor.gif", 
            "dims": [
              416, 
              224
            ], 
            "preview": "https://media.tenor.com/images/2a14201ce0f5aeaa27b0a1803255e9fd/tenor.gif", 
            "size": 1981318
          }, 
          "loopedmp4": {
            "url": "https://media.tenor.com/videos/afb6f589849a2dc12a7af58905f222b1/mp4", 
            "dims": [
              416, 
              224
            ], 
            "duration": 26.801302, 
            "preview": "https://media.tenor.com/images/1b7446613a50219636b4299992e6daf0/tenor.png", 
            "size": 1188713
          }
        }
      ], 
      "tags": [], 
      "shares": 1, 
      "itemurl": "https://tenor.com/view/office-space-dancing-break-dance-celebrate-gangsta-gif-11313704", 
      "composite": null, 
      "hasaudio": false, 
      "title": "", 
      "id": "11313704"
    }
  ], 
  "next": "0"
}""")


empty_gif_response = json.loads("""{
  "results": [], 
  "next": "0"
}""")


"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import io
import json
import unittest
from unittest.mock import patch, MagicMock, Mock

from PIL import Image

from gif_for_cli.generate import (
    _run_ffmpeg,
    _save_config,
    convert_frame,
    _convert_frames,
    generate,
)


@patch('gif_for_cli.generate.open')
class TestSaveConfig(unittest.TestCase):
    def test(self, mocked_open):
        num_frames = 10
        seconds = 1.5
        options = {
            'input_source': 'happy birthday',
            'input_source_file': 'foo.jpg',
            'cols': 160,
            'cell_width': 3,
            'cell_height': 6,
            'output_dirnames': {'.': 'foo'},
        }

        f = io.StringIO()
        # we need close() to noop so we can verify the value later
        f.close = lambda *args, **kwargs: None
        mocked_open.return_value = f

        _save_config(num_frames, seconds, **options)

        content = json.loads(f.getvalue())

        self.assertNotIn('output_dirnames', content)
        self.assertEqual(content, {
            'input_source': options['input_source'],
            'input_source_file': options['input_source_file'],
            'cols': options['cols'],
            'cell_width': options['cell_width'],
            'cell_height': options['cell_height'],
            'num_frames': num_frames,
            'seconds': seconds,
        })


@patch('gif_for_cli.generate.subprocess.Popen')
class TestRunFfmpeg(unittest.TestCase):
    def test(self, mock_Popen):
        out = b''
        err = b"""Output #0, image2, to '/home/foo/.cache/gif-for-cli/0.0.0/2094cb18c10ddb47dbe239ddbd702cc0-160cols-cw3px-ch6px/jpg/%04d.jpg':
  Metadata:
    major_brand     : isom
    minor_version   : 512
    compatible_brands: isomiso2avc1mp41
    encoder         : Lavf57.83.100
    Stream #0:0(und): Video: mjpeg, yuvj420p(pc), 480x257, q=2-31, 200 kb/s, 10 fps, 10 tbn, 10 tbc (default)
    Metadata:
      handler_name    : VideoHandler
      encoder         : Lavc57.107.100 mjpeg
    Side data:
      cpb: bitrate max/min/avg: 0/0/200000 buffer size: 0 vbv_delay: -1
frame=   11 fps=0.0 q=20.2 Lsize=N/A time=00:00:01.10 bitrate=N/A speed=3.21x    
video:258kB audio:0kB subtitle:0kB other streams:0kB global headers:0kB muxing overhead: unknown"""

        mock_process = Mock()
        mock_process.communicate.return_value = (out, err,)
        mock_Popen.return_value = mock_process

        options = {
            'input_source_file': 'foo.gif',
            'output_dirnames': {'jpg': 'foo/jpg'},
            'cols': 160,
            'rows': 160,
            'cell_width': 3,
            'cell_height': 6,
            'ignoreme': None,
        }

        num_frames, seconds = _run_ffmpeg(**options)

        self.assertEqual(num_frames, 11)
        self.assertEqual(seconds, 1.1)


@patch('gif_for_cli.generate.Image')
class TestConvertFrame(unittest.TestCase):
    def test(self, mock_Image):
        im = Image.new('RGB', (100, 100,))
        # this exercises some code branches that handle multiple colors
        for i in range(0, 10):
            im.putpixel((i, 0,), (255, 255, 5 * i,))

        mock_Image.open.return_value = im

        frame_name = '0001'
        options = {
            'cell_height': 6,
            'cell_width': 3,
            'output_dirnames': {
                'jpg': 'foo/jpg',
                'nocolor': 'foo/nocolor',
                '256': 'foo/256',
                '256fgbg': 'foo/256fgbg',
                'truecolor': 'foo/truecolor',
            },
        }

        with patch('gif_for_cli.generate.open') as mocked_open:
            convert_frame(frame_name, **options)

        self.assertEqual(mock_Image.open.call_count, 1)
        self.assertEqual(mocked_open.call_count, 4)


@patch('gif_for_cli.generate.get_sorted_filenames')
@patch('gif_for_cli.generate.convert_frame')
@patch('gif_for_cli.generate.Pool')
class TestConvertFrames(unittest.TestCase):
    def test_1_cpu(self, mock_Pool, mock_convert_frame, mock_get_sorted_filenames):
        mock_get_sorted_filenames.return_value = ['0001.jpg', '0002.jpg']

        options = {
            'cpu_pool_size': 1,
            'stdout': io.StringIO(),
            'output_dirnames': {
                'jpg': 'foo/jpg',
            },
        }

        _convert_frames(**options)

        self.assertEqual(mock_get_sorted_filenames.call_count, 1)
        self.assertEqual(mock_convert_frame.call_count, 2)
        self.assertEqual(mock_Pool.call_count, 0)

        output = options['stdout'].getvalue()

        self.assertEqual(output[-1], '\n')

        output = output[:-1].split(u'\u001b[2K\u001b[1000D')
        
        self.assertEqual(output, [
            'Processed 0/2 frames...',
            'Processed 1/2 frames...',
            'Processed 2/2 frames...',
        ])

    def test_2_cpus(self, mock_Pool, mock_convert_frame, mock_get_sorted_filenames):
        mock_get_sorted_filenames.return_value = ['0001.jpg', '0002.jpg']
        mock_pool = MagicMock()
        mock_pool.__enter__.return_value = mock_pool
        def mock_result(f, args, kwargs):
            m = Mock()
            m.get.return_value = f(*args, **kwargs)
            return m
        mock_pool.apply_async = mock_result
        mock_Pool.return_value = mock_pool

        options = {
            'cpu_pool_size': 2,
            'stdout': io.StringIO(),
            'output_dirnames': {
                'jpg': 'foo/jpg',
            },
        }

        _convert_frames(**options)

        self.assertEqual(mock_get_sorted_filenames.call_count, 1)
        self.assertEqual(mock_convert_frame.call_count, 2)
        self.assertEqual(mock_Pool.call_count, 1)

        output = options['stdout'].getvalue()

        self.assertEqual(output[-1], '\n')

        output = output[:-1].split(u'\u001b[2K\u001b[1000D')
        
        self.assertEqual(output, [
            'Processed 0/2 frames...',
            'Processed 1/2 frames...',
            'Processed 2/2 frames...',
        ])


@patch('gif_for_cli.generate._run_ffmpeg')
@patch('gif_for_cli.generate._save_config')
@patch('gif_for_cli.generate._convert_frames')
class TestGenerate(unittest.TestCase):
    def test(self, mock_convert_frames, mock_save_config, mock_run_ffmpeg):
        mock_run_ffmpeg.return_value = (11, 1.1,)

        options = {}

        generate(**options)

        self.assertEqual(mock_convert_frames.call_count, 1)
        self.assertEqual(mock_save_config.call_count, 1)
        self.assertEqual(mock_save_config.call_args[0][0], 11)
        self.assertEqual(mock_save_config.call_args[0][1], 1.1)
        self.assertEqual(mock_run_ffmpeg.call_count, 1)



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from collections import OrderedDict
import json
import math
from multiprocessing import Pool
import os
import re
import subprocess

from PIL import Image

from ..constants import ANSI_RESET, NOCOLOR_CHARS
from ..utils import get_sorted_filenames

from .utils import (
    avg,
    get_gray,
    get_256_cell,
    get_256fgbg_cell,
    get_truecolor_cell,
    get_avg_for_em,
    _log_frame_progress,
)


def _save_config(num_frames, seconds, **options):
    d = {
        key: options.get(key)
        for key in [
            'input_source',
            'input_source_file',
            'cols',
            'cell_width',
            'cell_height',
        ]
    }
    d['num_frames'] = num_frames
    d['seconds'] = seconds

    with open('{}/config.json'.format(options['output_dirnames']['.']), 'w') as f:
        json.dump(d, f)


def _run_ffmpeg(input_source_file, output_dirnames, cols, rows, cell_width, cell_height, **options):
    scale_width = cols * cell_width
    scale_height = rows * cell_height

    cmd = [
        'ffmpeg',
        '-i', input_source_file,
        '-vf', 'scale=w={}:h={}:force_original_aspect_ratio=decrease'.format(scale_width, scale_height),
        '{}/%04d.jpg'.format(output_dirnames['jpg'])
    ]
    p = subprocess.Popen(cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    out, err = p.communicate()
    err = err.decode('utf8')

    num_frames = int(re.search(r'frame=\s*(\d+)', err).group(1))
    hours, minutes, seconds = re.search(r'time=(\d{2}):(\d{2}):(\d{2}.\d{2})', err).groups()
    seconds = float(seconds) + (int(minutes) * 60) + (int(hours) * 3600)
    return num_frames, seconds


def convert_frame(frame_name, **options):
    cell_height = options['cell_height']
    cell_width = options['cell_width']
    output_dirnames = options['output_dirnames']

    # for each frame, generate text file with ANSI colors
    img = Image.open('{}/{}.jpg'.format(output_dirnames['jpg'], frame_name))
    px = img.load()
    width, height = img.size
    # trim image if needed.
    width = width - (width % cell_width)
    height = height - (height % cell_height)
    cols = math.floor(width / cell_width)

    chars_nocolor = []
    lines_256 = []
    lines_256fgbg = []
    lines_truecolor = []

    for y in range(0, height, cell_height):
        line_256 = []
        line_256fgbg = []
        line_truecolor = []
        for x in range(0, width, cell_width):
            rgb = get_avg_for_em(px, x, y, cell_height, cell_width)

            chars_nocolor.append(get_gray(*rgb))
            line_256.append(get_256_cell(*rgb))
            line_256fgbg.append(get_256fgbg_cell(*rgb))
            line_truecolor.append(get_truecolor_cell(*rgb))

        lines_256.append(''.join(line_256))
        # This output mode can leak the BG color to extra columns.
        lines_256fgbg.append(''.join(line_256fgbg) + ANSI_RESET)
        lines_truecolor.append(''.join(line_truecolor))

    # We need to divide up the gray colors into roughly equal buckets,
    # without adding numpy as a dependency just for the histogram function.
    num_cells_per_char = len(chars_nocolor) / len(NOCOLOR_CHARS)
    char_counts = OrderedDict()
    for cell in sorted(chars_nocolor):
        char_counts[cell] = char_counts.get(cell, 0) + 1
    cur_count = 0
    cur_char_idx = 0
    char_idxs = {}
    for cell, cell_num in char_counts.items():
        if cur_count > num_cells_per_char:
            cur_count = 0
            cur_char_idx += 1
        char_idxs[cell] = cur_char_idx
        cur_count += cell_num

    lines_nocolor = []
    line_nocolor = None
    for i, gray in enumerate(chars_nocolor):
        if not line_nocolor:
            line_nocolor = []

        line_nocolor.append(NOCOLOR_CHARS[char_idxs[gray]])

        if (i + 1) % cols == 0:
            lines_nocolor.append(''.join(line_nocolor))
            line_nocolor = None


    with open('{}/{}.txt'.format(output_dirnames['nocolor'], frame_name), 'w') as f:
        f.write('\n'.join(lines_nocolor))

    with open('{}/{}.txt'.format(output_dirnames['256'], frame_name), 'w') as f:
        f.write('\n'.join(lines_256))

    with open('{}/{}.txt'.format(output_dirnames['256fgbg'], frame_name), 'w') as f:
        f.write('\n'.join(lines_256fgbg))
    
    with open('{}/{}.txt'.format(output_dirnames['truecolor'], frame_name), 'w') as f:
        f.write('\n'.join(lines_truecolor))


def _convert_frames(cpu_pool_size, stdout, **options):
    output_dirnames = options['output_dirnames']

    frame_names = [
        filename.split('.')[0]
        for filename in get_sorted_filenames(output_dirnames['jpg'], 'jpg')
    ]

    total = len(frame_names)

    if cpu_pool_size == 1:
        results = (
            convert_frame(frame_name, **options)
            for frame_name in frame_names
        )
        _log_frame_progress(total, results, stdout)
    else:
        with Pool(cpu_pool_size) as pool:
            # we need this consumed instantly in order for the tasks to begin
            # execution in parallel.
            results = [
                pool.apply_async(convert_frame, [frame_name], options)
                for frame_name in frame_names
            ]
            # then use a generator to iterate as they execute.
            results = (r.get() for r in results)
            _log_frame_progress(total, results, stdout)


def generate(**options):
    # extract frames to files
    num_frames, seconds = _run_ffmpeg(**options)

    _save_config(num_frames, seconds, **options)

    _convert_frames(**options)



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import argparse
import unittest
from unittest.mock import patch, Mock

from gif_for_cli.utils import (
    _get_default_display_mode,
    _pool_type,
    get_parser,
    get_output_dirnames,
    get_sorted_filenames
)


class TestGetDefaultDisplayMode(unittest.TestCase):
    def test_empty_env(self):
        self.assertEqual(_get_default_display_mode({}), 'nocolor')

    def test_256(self):
        self.assertEqual(_get_default_display_mode({
            'TERM': 'xterm-256color',
        }), '256fgbg')

    def test_truecolor(self):
        self.assertEqual(_get_default_display_mode({
            'TERM': 'xterm-256color',
            'COLORTERM': 'truecolor',
        }), 'truecolor')


class TestPoolType(unittest.TestCase):
    def test_none(self):
        self.assertIsNone(_pool_type(None))

    def test_int_string(self):
        self.assertEqual(_pool_type('2'), 2)

    def test_int_string_0(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            _pool_type('0')


class TestGetParser(unittest.TestCase):
    def test(self):
        parser = get_parser({})

        self.assertIsNotNone(parser)


class TestGetOutputDirnames(unittest.TestCase):
    def test(self):
        output_dirnames = get_output_dirnames(
            '/home/foo',
            '0.0.0',
            '2094cb18c10ddb47dbe239ddbd702cc0',
            160,
            140,
            3,
            6
        )

        dirname = '/home/foo/.cache/gif-for-cli/0.0.0/2094cb18c10ddb47dbe239ddbd702cc0-160cols-140rows-cw3px-ch6px'

        self.assertEqual(output_dirnames['.'], dirname)
        self.assertEqual(output_dirnames['jpg'], dirname + '/jpg')
        self.assertEqual(output_dirnames['nocolor'], dirname + '/nocolor')
        self.assertEqual(output_dirnames['256'], dirname + '/256')
        self.assertEqual(output_dirnames['truecolor'], dirname + '/truecolor')


@patch('os.scandir')
class TestGetSortedFilenames(unittest.TestCase):
    def test(self, mock_scandir):
        display_dirname = 'some-dir'

        def mock_entry(name, is_file):
            m = Mock()
            m.name = name
            m.is_file.return_value = is_file
            return m

        mock_scandir.return_value = [
            mock_entry('.', False),
            mock_entry('..', False),
            mock_entry('sike.txt', False),
            mock_entry('.DS_Store', True),
            mock_entry('0001.txt', True),
            mock_entry('0002.txt', True),
        ]

        sorted_filenames = get_sorted_filenames(display_dirname, 'txt')

        self.assertEqual(mock_scandir.call_count, 1)
        self.assertEqual(mock_scandir.call_args[0][0], display_dirname)
        self.assertEqual(list(sorted_filenames), [
            '0001.txt',
            '0002.txt',
        ])



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import hashlib
import json
import os
from os.path import expanduser
import sys

from . import __version__
from .constants import STORED_CELL_CHAR
from .display import display
from .generate import generate
from .generate.utils import process_input_source
from .utils import get_parser, get_output_dirnames


def execute(environ, argv, stdout):
    parser = get_parser(environ)

    args = parser.parse_args(argv)


    input_source = args.input_source

    input_source_file = process_input_source(input_source, args.api_key)


    home_dir = expanduser('~')

    m = hashlib.md5()
    m.update(input_source_file.encode('utf8'))
    input_source_hash = m.hexdigest()

    output_dirnames = get_output_dirnames(
        home_dir,
        __version__,
        input_source_hash,
        args.cols,
        args.rows,
        args.cell_width,
        args.cell_height
    )

    if not os.path.exists(output_dirnames['.']):
        for output_dirname in output_dirnames.values():
            if not os.path.exists(output_dirname):
                os.makedirs(output_dirname)

        generate(
            stdout=stdout,
            input_source=input_source,
            input_source_file=input_source_file,
            cols=args.cols,
            rows=args.rows,
            cell_width=args.cell_width,
            cell_height=args.cell_height,
            output_dirnames=output_dirnames,
            cpu_pool_size=args.cpu_pool_size,
        )


    with open('{}/config.json'.format(output_dirnames['.']), 'r') as f:
        config = json.load(f)


    display(
        display_dirname=output_dirnames[args.display_mode],
        stdout=stdout,
        num_loops=args.num_loops,
        cell_char=args.cell_char,
        seconds_per_frame=config['seconds'] / config['num_frames']
    )


def main(): # pragma: no cover
    execute(os.environ, sys.argv[1:], sys.stdout)


if __name__ == '__main__': # pragma: no cover
    main()



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import itertools
from json.decoder import JSONDecodeError
import math
import os
import urllib.parse

import requests
from x256 import x256

from ..constants import X256FGBG_CHARS, STORED_CELL_CHAR

from .x256fgbg_utils import top_2_colors


def memoize(f):
    """
    Caveat: Presumes each arg is hashable, and therefore a valid dict key.
    """
    d = {}
    def wrapper(*args):
        if args in d:
            return d[args]

        res = f(*args)
        d[args] = res
        return res

    return wrapper


def avg(it):
    l = len(it)
    return sum(it) / float(l)


@memoize
def get_gray(*rgb):
    return avg(rgb)


@memoize
def get_256_cell(r, g, b):
    return u'\u001b[38;5;{}m{}'.format(x256.from_rgb(r, g, b), STORED_CELL_CHAR)


@memoize
def get_256fgbg_cell(r, g, b):
    best, second = top_2_colors(r, g, b)
    # if the best color is an exact match, use a blank space for the FG color.
    char = ' '
    if best['distance'] != 0:
        # The bigger the distance of the best color, the more we want the FG
        # color to show up.
        # Cases:
        # best and second are equal - 50BG/50FG - largest X256FGBG_CHARS
        # best is close, second is far away - 90BG/10FG - smaller X256FGBG_CHARS
        ratio = best['distance'] / second['distance']
        char = X256FGBG_CHARS[math.floor(ratio * (len(X256FGBG_CHARS) - 1))]

    return u'\u001b[48;5;{}m\u001b[38;5;{}m{}'.format(
        best['index'],
        second['index'],
        char,
    )


@memoize
def get_truecolor_cell(r, g, b):
    return u'\u001b[38;2;{};{};{}m{}'.format(r, g, b, STORED_CELL_CHAR)


def get_avg_for_em(px, x, y, cell_height, cell_width):
    pixels = [
        px[sx, sy]
        for sy in range(y, y + cell_height)
        for sx in range(x, x + cell_width)
    ]
    return [round(n) for n in map(avg, zip(*pixels))]


def process_input_source(input_source, api_key):
    if input_source.strip().startswith('https://tenor.com/view/'):
        gif_id = input_source.rsplit('-', 1)[-1]
        if gif_id.isdigit():
            input_source = gif_id
        else:
            raise Exception('Bad GIF URL.')

    is_url = input_source.startswith('http://') or input_source.startswith('https://')

    if not os.path.exists(input_source) and not is_url:
        # get from Tenor GIF API
        if input_source.isdigit():
            endpoint = 'gifs'
            query = 'ids={}'.format(urllib.parse.quote_plus(input_source))
        elif input_source == '':
            endpoint = 'trending'
            query = 'limit=1'
        else:
            endpoint = 'search'
            query = 'limit=1&q={}'.format(urllib.parse.quote_plus(input_source))

        resp = requests.get('https://api.tenor.com/v1/{}?key={}&{}'.format(endpoint, api_key, query))

        try:
            resp_json = resp.json()
        except JSONDecodeError:
            raise Exception('A server error occurred.')

        if 'error' in resp_json:
            raise Exception('An error occurred: {}'.format(resp_json['error']))

        results = resp_json.get('results')

        if not results:
            raise Exception('Could not find GIF.')

        input_source = results[0]['media'][0]['mp4']['url']
    return input_source


def _log_frame_progress(total, results, stdout):
    for count, result in enumerate(itertools.chain([None], results)):
        if count:
            stdout.write(u'\u001b[2K\u001b[1000D')
        stdout.write('Processed {}/{} frames...'.format(count, total))
        stdout.flush()
    stdout.write('\n')



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import argparse
import os


def _get_default_display_mode(environ):
    TERM = environ.get('TERM', '').lower()
    # FIXME: COLORTERM may not be accepted by sshd
    COLORTERM = environ.get('COLORTERM', '').lower()
    if 'truecolor' in TERM or 'truecolor' in COLORTERM:
        return 'truecolor'
    elif '256' in TERM or '256' in COLORTERM:
        return '256fgbg'
    return 'nocolor'


def _pool_type(val):
    if val is None:
        return val
    val = int(val)
    if val <= 0:
        raise argparse.ArgumentTypeError('Minimum cpu_pool_size is 1')
    return val


def get_parser(environ):
    default_display_mode = _get_default_display_mode(environ)

    parser = argparse.ArgumentParser(
        prog='gif_for_cli',
        description="""Convert .gif/.mp4 to animated ASCII art with or wihtout ANSI colors
    and view it in your terminal. Supports querying Tenor GIF API.
    """,
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        'input_source',
        type=str,
        nargs='?',
        default='',
        help="""An input source. Can be a filename, url, or a query to the Tenor GIF API.

    e.g.
    foo.gif - a local file.
    https://example.com/media.mp4 - a remote file.
    'happy birthday' - searches Tenor for "happy birthday" and gets the first result.
    12345 - gets the GIF with ID 12345 from Tenor.
    [none provided] - gets the first GIF from the Tenor trending endpoint.
    """
    )
    parser.add_argument(
        '--api-key',
        dest='api_key',
        type=str,
        default='TQ7VXFHXBJQ5',
        help='Tenor API key to perform queries with.'
    )
    # display related options.
    parser.add_argument(
        '--display-mode',
        dest='display_mode',
        type=str,
        default=default_display_mode,
        choices=['nocolor', '256', '256fgbg', 'truecolor'],
        help='Override the auto-detected color support.'
    )
    parser.add_argument(
        '-c',
        dest='cell_char',
        type=str,
        default='#',
        help='Character to use for each colorized cell/block. e.g. #, \u2588, etc.'
    )
    parser.add_argument(
        '-l',
        dest='num_loops',
        type=int,
        default=3,
        help='Number of times to repeat animation. 0 will repeat forever.'
    )
    # generation related options.
    parser.add_argument(
        '--cols',
        dest='cols',
        type=int,
        default=160,
        help='Maximum number of columns.'
    )
    parser.add_argument(
        '--rows',
        dest='rows',
        type=int,
        default=40,
        help='Maximum number of rows.'
    )
    parser.add_argument(
        '-cw',
        dest='cell_width',
        type=int,
        default=3,
        help='Number of pixels in width you want mapped to a single character.'
    )
    parser.add_argument(
        '-ch',
        dest='cell_height',
        type=int,
        default=6,
        help='Number of pixels in height you want mapped to a single character.'
    )

    # generation related options, but doens't affect generated output.
    parser.add_argument(
        '--pool-size',
        dest='cpu_pool_size',
        type=_pool_type,
        default=None
    )

    return parser


def get_output_dirnames(home_dir, version, input_source_hash, cols, rows, cell_width, cell_height):
    # include generator options in path
    output_dirnames = {
        '.': '{}/.cache/gif-for-cli/{}/{}-{}cols-{}rows-cw{}px-ch{}px'.format(
            home_dir,
            version,
            input_source_hash,
            cols,
            rows,
            cell_width,
            cell_height
        ),
    }
    output_dirnames['jpg'] = '{}/jpg'.format(output_dirnames['.'])
    output_dirnames['nocolor'] = '{}/nocolor'.format(output_dirnames['.'])
    output_dirnames['256'] = '{}/256'.format(output_dirnames['.'])
    output_dirnames['256fgbg'] = '{}/256fgbg'.format(output_dirnames['.'])
    output_dirnames['truecolor'] = '{}/truecolor'.format(output_dirnames['.'])
    return output_dirnames


def get_sorted_filenames(dirname, ext):
    return (
        de.name
        for de in sorted(os.scandir(dirname), key=lambda de: de.name)
        if de.is_file() and de.name.endswith('.' + ext)
    )



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import io
import json
import unittest
from unittest.mock import patch, Mock

from PIL import Image

from gif_for_cli.generate.utils import (
    avg,
    get_gray,
    get_256_cell,
    get_truecolor_cell,
    get_avg_for_em,
    process_input_source,
    _log_frame_progress,
)

from ..fixtures import empty_gif_response, gif_response


api_key = 'TQ7VXFHXBJQ5'


class TestAvg(unittest.TestCase):
    def test_empty_input_raises_error(self):
        with self.assertRaises(ZeroDivisionError):
            avg([])

    def test_single_int_in_float_out(self):
        out = avg([1])

        self.assertEqual(out, 1.0)
        self.assertEqual(type(out), float)

    def test_ints_in_float_out(self):
        out = avg([1, 2])

        self.assertEqual(out, 1.5)
        self.assertEqual(type(out), float)

    def test_floats_in_float_out(self):
        out = avg([1.1, 2.2])

        self.assertAlmostEqual(out, 1.65)
        self.assertEqual(type(out), float)


class TestGetGray(unittest.TestCase):
    def test(self):
        self.assertAlmostEqual(
            get_gray(0, 128, 255),
            127.66666666
        )


class TestGet256Cell(unittest.TestCase):
    def test(self):
        self.assertEqual(
            get_256_cell(0, 128, 255),
            u'\u001b[38;5;33m#'
        )


class TestGetTruecolorCell(unittest.TestCase):
    def test(self):
        self.assertEqual(
            get_truecolor_cell(0, 128, 255),
            u'\u001b[38;2;0;128;255m#'
        )


class TestGetAvgForEm(unittest.TestCase):
    def setUp(self):
        self.im = Image.new('RGB', (100, 100,))
        self.px = self.im.load()
        self.black = [0, 0, 0,]
        self.white = [255, 255, 255,]
        self.gray = [128, 128, 128,]

    def assertColor(self, out, color):
        self.assertEqual(out, color)

        for v in out:
            self.assertEqual(type(v), int)

    def test_default_black_block(self):
        out = get_avg_for_em(self.px, 0, 0, 2, 2)

        out = get_avg_for_em(self.px, 0, 0, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 2, 0, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 0, 2, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 2, 2, 2, 2)

        self.assertColor(out, self.black)

    def test_white_block(self):
        self.im.putpixel((0, 0,), tuple(self.white))
        self.im.putpixel((1, 0,), tuple(self.white))
        self.im.putpixel((0, 1,), tuple(self.white))
        self.im.putpixel((1, 1,), tuple(self.white))

        out = get_avg_for_em(self.px, 0, 0, 2, 2)

        self.assertColor(out, self.white)

        out = get_avg_for_em(self.px, 2, 0, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 0, 2, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 2, 2, 2, 2)

        self.assertColor(out, self.black)

    def test_gray_block(self):
        self.im.putpixel((0, 0,), tuple(self.white))
        self.im.putpixel((0, 1,), tuple(self.white))

        out = get_avg_for_em(self.px, 0, 0, 2, 2)

        self.assertColor(out, self.gray)

        out = get_avg_for_em(self.px, 2, 0, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 0, 2, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 2, 2, 2, 2)

        self.assertColor(out, self.black)

    def test_separate_color_channels(self):
        color1 = (0, 128, 255,)
        color2 = (255, 128, 0,)

        self.im.putpixel((0, 0,), color1)
        self.im.putpixel((1, 0,), color1)
        self.im.putpixel((0, 1,), color1)
        self.im.putpixel((1, 1,), color1)

        self.im.putpixel((2, 0,), color1)
        self.im.putpixel((3, 0,), color1)
        self.im.putpixel((2, 1,), color2)
        self.im.putpixel((3, 1,), color2)

        out = get_avg_for_em(self.px, 0, 0, 2, 2)

        self.assertColor(out, list(color1))

        out = get_avg_for_em(self.px, 2, 0, 2, 2)

        self.assertColor(out, self.gray)

        out = get_avg_for_em(self.px, 0, 2, 2, 2)

        self.assertColor(out, self.black)

        out = get_avg_for_em(self.px, 2, 2, 2, 2)

        self.assertColor(out, self.black)


@patch('os.path.exists')
@patch('gif_for_cli.generate.utils.requests')
class TestProcessInputSource(unittest.TestCase):
    def set_mock_response(self, mock_requests, data, side_effect=False):
        mock_response = Mock()
        if side_effect:
            mock_response.json.side_effect = data
        else:
            mock_response.json.return_value = data
        mock_requests.get.return_value = mock_response

    def test_file(self, mock_requests, mock_exists):
        mock_exists.return_value = True

        input_source = 'foo.gif'

        processed_input_source = process_input_source(input_source, api_key)

        self.assertEqual(processed_input_source, input_source)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 0)

    def test_http_url(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        input_source = 'http://example.com/foo.gif'

        processed_input_source = process_input_source(input_source, api_key)

        self.assertEqual(processed_input_source, input_source)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 0)

    def test_https_url(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        input_source = 'https://example.com/foo.gif'

        processed_input_source = process_input_source(input_source, api_key)

        self.assertEqual(processed_input_source, input_source)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 0)

    def test_tenor_trending(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, gif_response)

        input_source = ''

        processed_input_source = process_input_source(input_source, api_key)

        mpr_url = gif_response['results'][0]['media'][0]['mp4']['url']
        self.assertEqual(processed_input_source, mpr_url)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_search(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, gif_response)

        input_source = 'happy birthday'

        processed_input_source = process_input_source(input_source, api_key)

        mpr_url = gif_response['results'][0]['media'][0]['mp4']['url']
        self.assertEqual(processed_input_source, mpr_url)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_search_empty_results(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, empty_gif_response)

        input_source = 'happy birthday'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'Could not find GIF.')
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_id(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, gif_response)

        input_source = '11313704'

        processed_input_source = process_input_source(input_source, api_key)

        mpr_url = gif_response['results'][0]['media'][0]['mp4']['url']
        self.assertEqual(processed_input_source, mpr_url)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_id_error_occurred(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, {'error': 'some error'})

        input_source = '11313704'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'An error occurred: some error')
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_id_empty_json(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, {})

        input_source = '11313704'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'Could not find GIF.')
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_id_exception_when_getting_json(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, Exception('some error'), side_effect=True)

        input_source = '11313704'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'some error')
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_id_json_decode_error(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, lambda *args: json.loads('<'), side_effect=True)

        input_source = '11313704'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'A server error occurred.')
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_gif_url(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, gif_response)

        input_source = 'https://tenor.com/view/the-matrix-gif-5437241'

        processed_input_source = process_input_source(input_source, api_key)

        mpr_url = gif_response['results'][0]['media'][0]['mp4']['url']
        self.assertEqual(processed_input_source, mpr_url)
        self.assertEqual(mock_exists.call_count, 1)
        self.assertEqual(mock_requests.get.call_count, 1)

    def test_tenor_broken_gif_url(self, mock_requests, mock_exists):
        mock_exists.return_value = False

        self.set_mock_response(mock_requests, gif_response)

        input_source = 'https://tenor.com/view/the-matrix-gif'

        with self.assertRaises(Exception) as cm:
            process_input_source(input_source, api_key)

        self.assertEqual(cm.exception.args[0], 'Bad GIF URL.')
        self.assertEqual(mock_exists.call_count, 0)
        self.assertEqual(mock_requests.get.call_count, 0)


class TestLogFrameProgress(unittest.TestCase):
    def test(self):
        total = 5
        results = range(0, total)
        stdout = io.StringIO()

        _log_frame_progress(total, results, stdout)

        output = stdout.getvalue()

        self.assertEqual(output[-1], '\n')

        output = output[:-1].split(u'\u001b[2K\u001b[1000D')
        
        self.assertEqual(output, [
            'Processed 0/5 frames...',
            'Processed 1/5 frames...',
            'Processed 2/5 frames...',
            'Processed 3/5 frames...',
            'Processed 4/5 frames...',
            'Processed 5/5 frames...',
        ])



from x256 import x256


# The max possible distance (i.e. between #000 and #fff)
_MAX_DISTANCE = 256 ** 2 * 3


def top_2_colors(r, g, b):
    c = [r, g, b]
    best = {'distance': _MAX_DISTANCE, 'index': 1}
    second = {'distance': _MAX_DISTANCE, 'index': 1}

    for index, item in enumerate(x256.colors):
        d = x256.__distance(item, c)
        if(d <= best['distance']):
            if best:
              second = best
            best = {'distance': d, 'index': index}

    return best, second



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import io
import json
import unittest
from unittest.mock import mock_open, patch, MagicMock, Mock

from gif_for_cli.__main__ import execute


@patch('gif_for_cli.__main__.process_input_source')
@patch('gif_for_cli.__main__.os.makedirs')
@patch('gif_for_cli.__main__.generate')
@patch('gif_for_cli.__main__.display')
class TestExecute(unittest.TestCase):
    def test_new(self, mock_display, mock_generate, mock_makedirs, mock_process_input_source):
        mock_process_input_source.side_effect = lambda input_source, api_key: input_source

        environ = {}
        argv = []
        stdout = io.StringIO()

        input_source = ''
        cols = 160
        rows = 40
        cell_width = 3
        cell_height = 6
        num_frames = 11
        seconds = 1.1

        with patch('gif_for_cli.__main__.open') as mocked_open:
            mocked_open.return_value = io.StringIO(json.dumps({
                'input_source': input_source,
                'input_source_file': input_source,
                'cols': cols,
                'rows': rows,
                'cell_width': cell_width,
                'cell_height': cell_height,
                'num_frames': num_frames,
                'seconds': seconds,
            }))

            with patch('gif_for_cli.__main__.os.path.exists') as mock_exists:
                mock_exists.return_value = False

                execute(environ, argv, stdout)

        self.assertEqual(mock_process_input_source.call_count, 1)
        # for some reaosn, this intercepts some locale laoding
        paths = sorted([
            call[0][0]
            for call in mock_exists.call_args_list[-6:]
        ])

        self.assertTrue(paths[0].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px'
        ))
        self.assertTrue(paths[1].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px/256'
        ))
        self.assertTrue(paths[2].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px/256fgbg'
        ))
        self.assertTrue(paths[3].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px/jpg'
        ))
        self.assertTrue(paths[4].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px/nocolor'
        ))
        self.assertTrue(paths[5].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px/truecolor'
        ))
        self.assertEqual(mock_makedirs.call_count, 6)
        self.assertEqual(mock_generate.call_count, 1)
        self.assertEqual(mocked_open.call_count, 1)
        self.assertEqual(mock_display.call_count, 1)

    def test_cached(self, mock_display, mock_generate, mock_makedirs, mock_process_input_source):
        mock_process_input_source.side_effect = lambda input_source, api_key: input_source

        environ = {}
        argv = []
        stdout = io.StringIO()

        input_source = ''
        cols = 160
        rows = 40
        cell_width = 3
        cell_height = 6
        num_frames = 11
        seconds = 1.1

        with patch('gif_for_cli.__main__.open') as mocked_open:
            mocked_open.return_value = io.StringIO(json.dumps({
                'input_source': input_source,
                'input_source_file': input_source,
                'cols': cols,
                'rows': rows,
                'cell_width': cell_width,
                'cell_height': cell_height,
                'num_frames': num_frames,
                'seconds': seconds,
            }))

            with patch('gif_for_cli.__main__.os.path.exists') as mock_exists:
                mock_exists.return_value = True

                execute(environ, argv, stdout)

        self.assertEqual(mock_process_input_source.call_count, 1)
        # for some reaosn, this intercepts some locale laoding
        paths = sorted([
            call[0][0]
            for call in mock_exists.call_args_list[-6:]
        ])
        self.assertTrue(paths[0].endswith(
            '/d41d8cd98f00b204e9800998ecf8427e-160cols-40rows-cw3px-ch6px'
        ))
        self.assertEqual(mock_makedirs.call_count, 0)
        self.assertEqual(mock_generate.call_count, 0)
        self.assertEqual(mocked_open.call_count, 1)
        self.assertEqual(mock_display.call_count, 1)
        self.assertEqual(mock_display.call_args[1]['seconds_per_frame'], 0.1)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from setuptools import setup, find_packages
import gif_for_cli

setup(
    name='gif-for-cli',
    version=gif_for_cli.__version__,
    description="Render an animated GIF to your command line terminal.",
    author='Seán Hayes',
    author_email='sth@google.com',
    url='https://github.com/google/gif-for-cli',
    keywords='gif cli',
    packages=find_packages(),
    entry_points = {
        'console_scripts': ['gif-for-cli=gif_for_cli.__main__:main'],
    },
    install_requires=[
        'Pillow>=5.1.0', # PIL Software License
        'requests>=2.18.4', # Apache License 2.0
        'x256>=0.0.3', # MIT License
    ],
    tests_require=[
        'coverage>=4.5.1',
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
        'Topic :: Artistic Software',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Multimedia :: Graphics :: Presentation',
        'Topic :: Multimedia :: Graphics :: Viewers',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Multimedia :: Video :: Display',
        'Topic :: Terminals',
        'Topic :: Utilities',
    ],
)



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import time

from .constants import STORED_CELL_CHAR, ANSI_RESET, ANSI_CURSOR_UP
from .utils import get_sorted_filenames


def display_txt_frames(txt_frames, stdout, num_loops, seconds_per_frame):
    previous_line_count = 0
    remaining_loops = num_loops or None

    try:
        while remaining_loops is None or remaining_loops > 0:
            for txt_frame in txt_frames:
                stdout.write(ANSI_CURSOR_UP * previous_line_count)
                stdout.write(txt_frame)
                stdout.write('\n')
                stdout.flush()
                previous_line_count = len(txt_frames[0].split('\n'))
                time.sleep(seconds_per_frame)

            if remaining_loops is not None:
                remaining_loops -= 1
        stdout.write(ANSI_RESET)
    except KeyboardInterrupt:
        # ensure styling is reset
        stdout.write(ANSI_RESET)
        # we'll want an extra new line if CTRL+C was pressed
        stdout.write('\n')

    stdout.flush()



def get_txt_frames(display_dirname, cell_char):
    return [
        open('{}/{}'.format(display_dirname, filename)).read().replace(STORED_CELL_CHAR, cell_char)
        for filename in get_sorted_filenames(display_dirname, 'txt')
    ]


def display(display_dirname, stdout, num_loops, cell_char, seconds_per_frame):
    txt_frames = get_txt_frames(display_dirname, cell_char)

    display_txt_frames(txt_frames, stdout, num_loops, seconds_per_frame)



"""
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import io
import unittest
from unittest.mock import patch

from gif_for_cli.constants import ANSI_CURSOR_UP, ANSI_RESET, STORED_CELL_CHAR
from gif_for_cli.display import display_txt_frames, get_txt_frames, display


class TestDisplayTxtFrames(unittest.TestCase):
    def setUp(self):
        self.num_frames = 5
        self.width = 10
        self.height = 5
        self.txt_frames = [
            '\n'.join([str(i) * self.width] * self.height)
            for i in range(self.num_frames)
        ]
        self.seconds_per_frame = 200

    def test_3_lopps(self):
        stdout = io.StringIO()

        txt_frames = self.txt_frames
        num_loops = 3

        with patch('time.sleep') as mock_sleep:
            display_txt_frames(txt_frames, stdout, num_loops, self.seconds_per_frame)

        self.assertEqual(mock_sleep.call_count, num_loops * len(txt_frames))
        for call in mock_sleep.call_args_list:
            self.assertEqual(call[0][0], self.seconds_per_frame)

        output_ending = '\n' + ANSI_RESET
        output = stdout.getvalue()

        self.assertEqual(output[-len(output_ending):], output_ending)

        output = output[:-len(output_ending)]
        output = output.split('\n' + (ANSI_CURSOR_UP * self.height))

        self.assertEqual(output, self.txt_frames * num_loops)

    def test_0_lopps(self):
        stdout = io.StringIO()

        txt_frames = self.txt_frames
        num_loops = 0
        error_after_num_loops = 5
        error_after_num_sleep_calls = error_after_num_loops * len(txt_frames)

        with patch('time.sleep') as mock_sleep:
            num_sleep_calls = 0
            def sleep_side_effect(s):
                nonlocal num_sleep_calls
                num_sleep_calls += 1
                if num_sleep_calls >= error_after_num_sleep_calls:
                    raise KeyboardInterrupt()
                return
            mock_sleep.side_effect = sleep_side_effect

            display_txt_frames(txt_frames, stdout, num_loops, self.seconds_per_frame)

        self.assertEqual(mock_sleep.call_count, error_after_num_loops * len(txt_frames))
        for call in mock_sleep.call_args_list:
            self.assertEqual(call[0][0], self.seconds_per_frame)

        output_ending = '\n' + ANSI_RESET + '\n'
        output = stdout.getvalue()

        self.assertEqual(output[-len(output_ending):], output_ending)

        output = output[:-len(output_ending)]
        output = output.split('\n' + (ANSI_CURSOR_UP * self.height))

        self.assertEqual(output, self.txt_frames * error_after_num_loops)


@patch('gif_for_cli.display.open')
@patch('gif_for_cli.display.get_sorted_filenames')
class TestGetTxtFrames(unittest.TestCase):
    def test(self, mock_get_sorted_filenames, mock_open):
        display_dirname = 'some-dir'
        cell_char = '$'

        mock_get_sorted_filenames.return_value = ['0001.txt', '0002.txt']
        mock_open.side_effect = lambda *args, **kwargs: io.StringIO(STORED_CELL_CHAR * 10)

        txt_frames = get_txt_frames(display_dirname, cell_char)

        self.assertEqual(len(txt_frames), 2)
        self.assertEqual(txt_frames[0], cell_char * 10)
        self.assertEqual(txt_frames[1], cell_char * 10)

        self.assertEqual(mock_get_sorted_filenames.call_count, 1)
        self.assertEqual(mock_get_sorted_filenames.call_args[0][0], display_dirname)
        self.assertEqual(mock_open.call_count, 2)
        self.assertEqual(mock_open.call_args_list[0][0][0], display_dirname + '/' + '0001.txt')
        self.assertEqual(mock_open.call_args_list[1][0][0], display_dirname + '/' + '0002.txt')


@patch('gif_for_cli.display.get_txt_frames')
@patch('gif_for_cli.display.display_txt_frames')
class TestDisplay(unittest.TestCase):
    def test(self, mock_display_txt_frames, mock_get_txt_frames):
        display_dirname = 'some-dir'
        stdout = io.StringIO()
        num_loops = 3
        cell_char = '$'
        seconds_per_frame = 0.1

        display(display_dirname, stdout, num_loops, cell_char, seconds_per_frame)

        self.assertEqual(mock_get_txt_frames.call_count, 1)
        self.assertEqual(mock_get_txt_frames.call_args[0][0], display_dirname)
        self.assertEqual(mock_get_txt_frames.call_args[0][1], cell_char)

        self.assertEqual(mock_display_txt_frames.call_count, 1)
        self.assertEqual(mock_display_txt_frames.call_args[0][0], mock_get_txt_frames.return_value)
        self.assertEqual(mock_display_txt_frames.call_args[0][1], stdout)
        self.assertEqual(mock_display_txt_frames.call_args[0][2], num_loops)
        self.assertEqual(mock_display_txt_frames.call_args[0][3], seconds_per_frame)
